/*
 * Definitions for treequery type
 * Internal structure as tree type, but element of array
 * can have several bits.
 * Teodor Sigaev <teodor@stack.net>
 */

#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include <ctype.h>

#include "entree.h"

PG_FUNCTION_INFO_V1(entreequery_in);
PG_FUNCTION_INFO_V1(entreequery_out);

Datum	entreequery_in(PG_FUNCTION_ARGS);
Datum	entreequery_out(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(entq_contains);
PG_FUNCTION_INFO_V1(entq_contain);

Datum   entq_contains(PG_FUNCTION_ARGS);
Datum   entq_contain(PG_FUNCTION_ARGS);

/* this function is used in gist, so declaration in entree.h */
PG_FUNCTION_INFO_V1(entree2query);

static int 
cmp_leveltype(const void* a, const void* b) {
	if ( *((leveltype*)a) == *((leveltype*)b) ) return 0;
	return (  *((leveltype*)a) > *((leveltype*)b) ) ? 1 : -1;
}

static void
prepare_lq(levelquery *lq) {
	register leveltype	*curend,*tmp;

	curend = tmp = lq->var;	
	qsort( (void*)tmp, lq->size, sizeof(leveltype), cmp_leveltype );

	/* unique */
	while (tmp - lq->var < lq->size)
		if (*tmp != *curend)
			*(++curend) = *tmp++;
		else
			tmp++;
	lq->size = curend + 1 - lq->var;
}

Datum 
entreequery_in(PG_FUNCTION_ARGS) {
	char    *buf = (char *) PG_GETARG_POINTER(0);
	int4	depth=1, palen, len=0, tmp, varlen=1;
	register char *ptr;
	treequery *node;
	levelquery **prsnode, *lqptr;
	register int4 state = PRS_WAITLEVEL, i=0;
	char numbuf[NUMLEN];

	ptr = buf;
	while( ptr && *ptr ) {
		if ( *ptr=='.' )
			depth++;
		else if ( *ptr=='|' )
			varlen++;
		ptr++;
	} 
	prsnode = (levelquery**)palloc( sizeof(levelquery*) * depth );

	ptr = buf;
	while( 1 ) {
		if ( state == PRS_WAITLEVEL ) {
			prsnode[len] = (levelquery*)palloc( LQ_CALCSIZE(varlen) );
			prsnode[len]->size=0;			
			if ( isdigit( (unsigned char) *ptr ) ) {
				state = PRS_WAITNUM;
				continue;
			} else if ( *ptr=='*' ) {
				state = PRS_WAITDELIM;
			} else {
				if ( *ptr )
					elog(ERROR, "Unknown character: '%c'", *ptr);
				else
					elog(ERROR, "Unexpected end of string");
			}
		} else if ( state == PRS_WAITNUM ) {
			if ( !isdigit( (unsigned char) *ptr ) ) {
				if ( *ptr )
					elog(ERROR, "Unknown character: '%c'", *ptr);
				else
					elog(ERROR, "Unexpected end of string");
			}
			numbuf[0] = *ptr; i=1;
			state = PRS_INNUM;
		} else if ( state == PRS_INNUM ) {
			if ( isdigit( (unsigned char) *ptr ) ) {
				if ( i>=NUMLEN-1 )
					elog(ERROR, "Number is too long");
				numbuf[i] = *ptr; i++;
			} else if ( *ptr == '.' || *ptr == '|' || !*ptr ) {
				numbuf[i] = '\0';
				tmp = atoi( numbuf );
				if ( tmp > MAXCHILD || tmp < 0 )
					elog(ERROR, "Invalid number of child: %d, ( 0<=n<=%d )", tmp, MAXCHILD);
				if ( 
						(tmp == 0 && prsnode[len]->size > 0) ||
						(tmp !=0  && prsnode[len]->size > 0 && prsnode[len]->var[ prsnode[len]->size-1 ]==0 )
				   	) 
						elog(ERROR, "Zero can't be used as alternative");
				prsnode[len]->var[ prsnode[len]->size ] = tmp;
				prsnode[len]->size++;	
				if ( *ptr == '|' ) {
					state = PRS_WAITNUM;
				} else if ( *ptr == '.' || !*ptr) {
					state = PRS_WAITDELIM;
					continue;
				} else
					elog(ERROR, "Inner logic error");
			} else 
				elog(ERROR, "Unknown character: '%c'", *ptr);
		} else if ( state == PRS_WAITDELIM ) {
			if ( *ptr == '.' ) {
				len++;
				state = PRS_WAITLEVEL;
			} else if ( !*ptr ) {
				len++;
				break;
			} else
				elog(ERROR, "Unknown character: '%c'", *ptr);
		} else
			elog(ERROR, "Inner logic error");
		ptr++; 
	}

	/* some shorter */
	while( len>0 && prsnode[len-1]->size == 0 )
		len--; 
	while( len>1 &&
			prsnode[len-1]->size && prsnode[len-1]->var[0] == (leveltype)0 && 
			prsnode[len-2]->size && prsnode[len-2]->var[0] == (leveltype)0 )
		len--;
	/* check */
	if ( prsnode[0]->var[0] == 0 ) 
		elog(ERROR, "Such treequery will produce always void result");
	for( i=0; i<len-1 ; i++ ) 
		if ( prsnode[len-2]->size && prsnode[i]->var[0] == 0 )
			elog(ERROR, "Zero can't be used in middle");

	/* size count */
	palen = TQ_HDRSIZE;
	for(i=0;i<len;i++) {
		if ( prsnode[i]->size > 1 )
			prepare_lq( prsnode[i] );
		palen += LQ_CALCSIZE( prsnode[i]->size );
	}
 
	node = (treequery*)palloc( palen );
	node->len = palen;
	node->size = len;
	lqptr = node->query;
	for(i=0;i<len;i++) {
		memcpy( (void*)lqptr, (void*)prsnode[i], LQ_CALCSIZE( prsnode[i]->size ) );
		pfree( prsnode[i] );
		lqptr = LQ_NEXT(lqptr);
	}
	pfree( prsnode );

	PG_RETURN_POINTER(node);
}

Datum
entreequery_out(PG_FUNCTION_ARGS) {
	treequery *node = (treequery*) PG_GETARG_POINTER(0);
	register int4 i,j;
	levelquery *lqptr = node->query;
	char *buf;
	register char *ptr;
	int first;

	/* palloc more than require */	
	ptr = buf = (char*)palloc( sizeof(char) * NUMLEN * (node->len - TQ_HDRSIZE) );
	*buf = '\0';
	for( i=0; i<node->size; i++ ) {
		first = 0;
		if ( lqptr->size == 0 ) {
			*ptr = '*';
			ptr++;
			*ptr = '\0';
		} else if ( lqptr->var[0] == 0 ) {
			*ptr = '0';
			ptr++;
			*ptr = '\0';
		} else 
			for(j=0; j< lqptr->size ; j++) {
				if ( j ) 
					sprintf(ptr,"|%d", lqptr->var[j] );
				else
					sprintf(ptr,"%d", lqptr->var[j] );
				ptr = strchr( ptr, '\0' );
			}
		
		if ( i!=node->size-1 ) {
			*ptr = '.';
			ptr++;
		}
		lqptr = LQ_NEXT(lqptr);
	}
	buf = (char*)repalloc( (void*)buf, sizeof(char) * ( ptr-buf+1 ) );
	PG_RETURN_POINTER(buf);
}

bool 
inner_entq_contain(entreenode *node, treequery *mask ) {
	register leveltype *s2 = node->id;
	int4 lenn = GETNODELEN(node);
	levelquery	*lqptr = mask->query;
	int len = min(lenn,mask->size);
	bool res=false;
	register leveltype *StopLow,*StopHigh,*StopMiddle;

	do {
		if ( lqptr->size==1 ) {
			if ( lqptr->var[0] != *s2 )
				return false;
		} else if ( lqptr->size ) {
			StopLow = lqptr->var;
			StopHigh = lqptr->var + lqptr->size;

			/* Loop invariant: StopLow <= val < StopHigh */
			res = true;
			while (StopLow < StopHigh) {
				StopMiddle = StopLow + (StopHigh - StopLow) / 2;
				if (*StopMiddle == *s2) {
					res = false;
					break;	
				} else if (*StopMiddle < *s2 )
					StopLow = StopMiddle + 1;
				else
					StopHigh = StopMiddle;
			}
			
			if ( res )
				return false; 

		}
		lqptr = LQ_NEXT(lqptr);
		s2++;
		len--;
	} while( len );

	if ( mask->size - 1 == lenn  && lqptr->var[0] == 0 ) 
		return true;
	return ( mask->size > lenn ) ? false : true;	
}

Datum
entq_contain(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		inner_entq_contain(
			(entreenode*) PG_GETARG_POINTER(0),
			(treequery*) PG_GETARG_POINTER(1)
		)
	);
}

Datum
entq_contains(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		inner_entq_contain(
			(entreenode*) PG_GETARG_POINTER(1),
			(treequery*) PG_GETARG_POINTER(0)
		)
	);
}

Datum
entree2query(PG_FUNCTION_ARGS) {
	entreenode *node = (entreenode*) PG_GETARG_POINTER(0);
	treequery *query;
	register levelquery      *lqptr;
	register int4 i, len = GETNODELEN(node);

	i = TQ_HDRSIZE + LQ_CALCSIZE(1)*len;
	query = (treequery*) palloc( i );
	query->len = i;
	query->size = len;
	lqptr = query->query;

	for(i=0;i<len;i++) {
		lqptr->size=1;
		lqptr->var[0] = node->id[i];
		lqptr = LQ_NEXT(lqptr);
	}

	PG_RETURN_POINTER(query);
}

