/*
 * GiST interface for bitree type
 * Teodor Sigaev <teodor@stack.net>
 */


#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include "access/nbtree.h"

#include "bitree.h"

/* used for sorting */
typedef struct rix
{
        int                     index;
        bitreenode       *r;
}       RIX;

PG_FUNCTION_INFO_V1( gbitree_compress );
Datum   gbitree_compress(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gbitree_consistent );
Datum   gbitree_consistent(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gbitree_union );
Datum   gbitree_union(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gbitree_same );
Datum   gbitree_same(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gbitree_penalty );
Datum   gbitree_penalty(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gbitree_picksplit );
Datum   gbitree_picksplit(PG_FUNCTION_ARGS);

Datum
gbitree_compress(PG_FUNCTION_ARGS) {
	PG_RETURN_DATUM( PG_GETARG_DATUM(0) );
}

static bool
cmp_le( bitreenode *innernode, bitreenode *node ) {
	int len1 = GETNODELEN(innernode);
	int len2 = GETNODELEN(node);
	register leveltype *ptr1 = innernode->id, *ptr2 = node->id;
	register leveltype min;

	while( ptr1 - innernode->id < len1 && ptr2 - node->id < len2 ) {
		if ( *ptr1 & *ptr2 ) {
			ptr1++; ptr2++;
			if ( *ptr1 != *ptr2 )
				return true; 
			continue;
		}
		min = 0;
		while ( (( *ptr1 >> min )&0x01) == 0 )
			min++;
		min = 0x01 << min;

		return ( min > *ptr2 ) ? false : true;
	}

	return true;
}

static bool
cmp_ge( bitreenode *innernode, bitreenode *node ) {
	int len1 = GETNODELEN(innernode);
	int len2 = GETNODELEN(node);
	register leveltype *ptr1 = innernode->id, *ptr2 = node->id;
	register leveltype max;

	while( ptr1 - innernode->id < len1 && ptr2 - node->id < len2 ) {
		if ( *ptr1 & *ptr2 ) {
			ptr1++; ptr2++;
			if ( *ptr1 != *ptr2 )
				return true; 
			continue;
		}
		max = MAXCHILD-1;
		while ( (( *ptr1 >> max )&0x01) == 0 )
			max--;
		max = 0x01 << max;

		return ( max < *ptr2 ) ? false : true;
	}

	return true;
}

static bool
g_is_parentin( bitreenode *key, bitreenode *query ) {
	int4    len = query->len;
	register int4 i;

	for( i=GETNODELEN(query); i>=0 ; i-- ) {  
		query->len = CALCSIZE(i); 
		if ( inner_bitq_contain( key, query, true ) ) {
			query->len = len;
			return true;
		}
	}
	query->len = len;

	return false;
}

Datum   
gbitree_consistent(PG_FUNCTION_ARGS) {
	GISTENTRY *entry = (GISTENTRY*)PG_GETARG_POINTER(0);
	bitreenode *query = (bitreenode*)PG_GETARG_POINTER(1);
	bitreenode *key = (bitreenode*)DatumGetPointer( entry->key );
	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
	bool res = false;

	switch( strategy ) {
		case BTLessStrategyNumber:
			res = ( GIST_LEAF( entry ) ) ?
					( cmp_binode(key,query) < 0 )
				:
					cmp_le(key, query);
			break;
		case BTLessEqualStrategyNumber:
			res = ( GIST_LEAF( entry ) ) ?
					( cmp_binode(key,query) <= 0 )
				:
					cmp_le(key, query);
			break;
		case BTEqualStrategyNumber:
			if ( GIST_LEAF( entry ) ) 
				res = ( cmp_binode(key,query) == 0 ) ? true : false;
			else
				res = inner_bitq_contain(key, query, true);
			break;
		case BTGreaterEqualStrategyNumber:
			res = ( GIST_LEAF( entry ) ) ?
					( cmp_binode(key,query) >= 0 )
				:
					cmp_ge(key, query);
			break;
		case BTGreaterStrategyNumber:
			res = ( GIST_LEAF( entry ) ) ?
					( cmp_binode(key,query) > 0 )
				:
					cmp_ge(key, query);
			break;
		case 6:
		case 7:
		case 11:
			res = inner_bitq_contain( key, query,
				!GIST_LEAF( entry ) );
			break;
		case 10:
			res = ( GIST_LEAF( entry ) ) ?
					inner_bi_isparent(key,query)
				:
					g_is_parentin(key,query);
			break;	
		default:
			elog(ERROR,"Unknown StrategyNumber: %d", strategy);
	}
	PG_RETURN_BOOL(res);
}

Datum
gbitree_union(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea *) PG_GETARG_POINTER(0);
	int *size = (int *) PG_GETARG_POINTER(1);
	int4 lenvec = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
	int4 lenunion=0,lentmp=0;
	register int4 i,j;
	bitreenode *result;
	leveltype *cur;
	int4 min,max;

	for(i=0;i<lenvec;i++)
		if ( (lentmp=GETNODELEN(GETENTRY(entryvec,i))) > lenunion )
			lenunion=lentmp;

	*size = CALCSIZE(lenunion);
	result = (bitreenode*)palloc( *size );
	MemSet( (void*)result,0,*size );
	result->len = *size;

	for(i=0;i<lenvec;i++) {
		cur = ( (bitreenode*)GETENTRY(entryvec,i) )->id;
		lentmp = GETNODELEN( GETENTRY(entryvec,i) );
		for(j=0;j<lentmp;j++)
			result->id[j] |= cur[j];
	}

	for(i=0;i<lentmp;i++) {
		min=MAXCHILD;
		max=0;
		for(j=0;j<MAXCHILD;j++) {
			if ( (0x01 & (result->id[i] >> j)) ) {
				max = j;
				if ( min==MAXCHILD )
					min=j;
			} 
		}
		for(j=min;j<=max;j++) 
			result->id[i] |= 0x01 << j;
	}

	PG_RETURN_POINTER( result );
}

Datum
gbitree_same(PG_FUNCTION_ARGS) {
	bool *result = (bool *)PG_GETARG_POINTER(2);

	*result = ( 
		cmp_binode( 
			(bitreenode*)PG_GETARG_POINTER(0), 
			(bitreenode*)PG_GETARG_POINTER(1) 
		) == 0 
	) ? true : false;
	PG_RETURN_POINTER(result); 
}

Datum
gbitree_penalty(PG_FUNCTION_ARGS) {
	bitreenode *origval = (bitreenode*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(0) )->key );
	bitreenode *newval  = (bitreenode*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(1) )->key );
	float  *penalty = (float *)    PG_GETARG_POINTER(2);
	register int4 i,j;
	int4 len,leno = GETNODELEN(origval), lenn = GETNODELEN( newval );
	int4 mino,maxo,minn,maxn;

	len = min(leno, lenn);
	*penalty = 0.0;
	for(i=0;i<len;i++) {
		minn=mino=MAXCHILD;
		maxn=maxo=0;
		for(j=0;j<MAXCHILD;j++) {
			if ( (0x01 & (origval->id[i] >> j)) ) {
				maxo=j;
				if ( mino==MAXCHILD )
					mino=j;
			}
			if ( (0x01 & (newval->id[i] >> j)) ) {
				maxn=j;
				if ( minn==MAXCHILD )
					minn=j;
			}
		}
		/* *penalty += ((float)(max( maxn-maxo, 0 ) + max( mino-minn, 0 ))) / ( (float)(i+1) ); */
		*penalty += (float)(max( maxn-maxo, 0 ) + max( mino-minn, 0 ));
	}

	PG_RETURN_POINTER(penalty);
}

static int
bitreekey_cmp(const void *a, const void *b) {
	register int i;
	register leveltype *sa,*sb;
	leveltype *ida, *idb;
	int4 lena = GETNODELEN(((RIX *) a)->r);
	int4 lenb = GETNODELEN(((RIX *) b)->r);
	
	sa = ida = (((RIX *) a)->r)->id;
	sb = idb = (((RIX *) b)->r)->id;

	while( sa-ida<lena && sb-idb<lenb ) {
		for(i=0;i<MAXCHILD;i++) {
			if ( (*sa >> i) & 0x01 ) {
				if ( (*sb >> i) & 0x01 )
					break;
				else
					return -1;
			} else if ( (*sb >> i) & 0x01 ) 
				return 1;
		}
		sa++; sb++;
	}
	return lena-lenb;
}

Datum
gbitree_picksplit(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea*) PG_GETARG_POINTER(0);
	GIST_SPLITVEC *v = (GIST_SPLITVEC*) PG_GETARG_POINTER(1);
	register OffsetNumber i;
	register int j;
	RIX *array;
	OffsetNumber maxoff;
	int	nbytes;
	int    maxlen = 0, size;
	bitreenode *lu,*ru;
	int min,max;

	maxoff = ((VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY)) - 1;
	nbytes = (maxoff + 2) * sizeof(OffsetNumber);
	v->spl_left = (OffsetNumber *) palloc(nbytes);
	v->spl_right = (OffsetNumber *) palloc(nbytes);
	v->spl_nleft = 0; 
	v->spl_nright = 0;
	array = (RIX *) palloc(sizeof(RIX) * (maxoff + 1));

	/* copy the data into RIXes, and sort the RIXes */
	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
		array[i].index = i;
		array[i].r = GETENTRY( entryvec, i );
		if ( GETNODELEN( array[i].r ) > maxlen )
			maxlen = GETNODELEN( array[i].r );
	}

	qsort((void *) &array[FirstOffsetNumber], maxoff - FirstOffsetNumber + 1,
		sizeof(RIX), bitreekey_cmp);

	size = CALCSIZE( maxlen );
	lu = (bitreenode*)palloc( size );
	ru = (bitreenode*)palloc( size );
	MemSet( (void*)lu->id,0,sizeof(leveltype)*maxlen);
	MemSet( (void*)ru->id,0,sizeof(leveltype)*maxlen);
	lu->len = ru->len = size;

	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
		if (i <= (maxoff - FirstOffsetNumber + 1) / 2) {
			v->spl_left[v->spl_nleft] = array[i].index;
			v->spl_nleft++;
			for(j=0;j<GETNODELEN(array[i].r);j++)
				lu->id[j] |= array[i].r->id[j];
		} else {
			v->spl_right[v->spl_nright] = array[i].index;
			v->spl_nright++;
			for(j=0;j<GETNODELEN(array[i].r);j++)
				ru->id[j] |= array[i].r->id[j];
		}
	}
	pfree(array);

	for(j=maxlen-1 ; j>=0 && lu->id[j]==0 ; j--);
	lu->len = CALCSIZE(j+1);
	for(i=0;i<GETNODELEN(lu);i++) {
		min=MAXCHILD;
		max=0;
		for(j=0;j<MAXCHILD;j++) {
			if ( (0x01 & (lu->id[i] >> j)) ) {
				max = j;
				if ( min==MAXCHILD )
					min=j;
			} 
		}
		for(j=min;j<=max;j++) 
			lu->id[i] |= 0x01 << j;
	}

	for(j=maxlen-1 ; j>=0 && ru->id[j]==0 ; j--);
	ru->len = CALCSIZE(j+1);
	for(i=0;i<GETNODELEN(ru);i++) {
		min=MAXCHILD;
		max=0;
		for(j=0;j<MAXCHILD;j++) {
			if ( (0x01 & (ru->id[i] >> j)) ) {
				max = j;
				if ( min==MAXCHILD )
					min=j;
			} 
		}
		for(j=min;j<=max;j++) 
			ru->id[i] |= 0x01 << j;
	}

	v->spl_ldatum = PointerGetDatum(lu);
	v->spl_rdatum = PointerGetDatum(ru);

	PG_RETURN_POINTER(v);
}

